from ._functions import auto
from ._functions import greater
from ._functions import lower

from ._classes import EncoderNone as none
from ._classes import EncoderDiscrete as discrete
from ._classes import EncoderIgnore as ignore
from ._classes import EncoderLimit as limit
from ._classes import EncoderOHE as OHE
from ._classes import EncoderEquivalence as Equivalence

__all__ = ['auto', 'greater', 'lower', 'none', 'discrete', 'ignore', 'limit', 'OHE', 'Equivalence']


from ._GMM import GMM
from ._KDE import KDE

def VAE(device='auto', **kwargs):
    # Avoid importing more packages than necessary when possible
    from ._VAE import VAE as _VAE
    return _VAE(device, **kwargs)

__all__ = ['GMM', 'KDE', 'VAE']


# -*- coding: utf-8 -*-
"""
Kernel Density Estimation

@author: david
"""

import numpy as np

from ._base import BaseGenerator

class KDE(BaseGenerator):
    def __init__(self, var=lambda n, d: np.power(n, -1 / d), **kwargs):
        super().__init__(**kwargs)
        if type(var) == type(lambda:0):
            self.var = var
        else:
            self.var = lambda n, d: var
        
    def probabilities(self, X):
        assert X.shape[1] == self.dim, "Size mismatch"
        dist = np.sum(np.square(
            np.reshape(self.X, (1, -1, self.dim))
            - np.reshape(X, (-1, 1, self.dim))
            ), 2)
        probs = np.sum(np.exp(-dist / 2 / self.var(self.n, self.dim)) / np.power(2 * np.pi * self.var(self.n, self.dim), self.dim / 2), 1) / self.n
        return probs
    
    def _fit(self, X):
        self.X = X
        self.n, self.dim = X.shape
            
    def _generate(self, size):
        ind = np.random.choice(self.n, size)
        S = np.random.normal(self.X[ind], np.sqrt(self.var(self.n, self.dim)))
        return S
    
    def _fill(self, Y):
        assert Y.shape[1] == self.dim, "Size mismatch"
        diffs = np.reshape(self.X, (1, -1, self.dim)) - np.reshape(Y, (-1, 1, self.dim))
        dists = np.sum(np.square(np.nan_to_num(diffs)), 2)
        probs = np.exp(-dists / 2 / self.var(self.n, self.dim))
        for y, prob in zip(Y, probs):
            bad = np.isnan(y)
            ind = np.random.choice(self.n, p=prob / np.sum(prob))
            y[bad] = np.random.normal(self.X[ind, bad], np.sqrt(self.var(self.n, self.dim)))
        return Y
    


# -*- coding: utf-8 -*-
"""
EMD

@author: david
"""

import numpy as np

def best(dist, move):
    is_best = True
    base = np.arange(move.size)
    for i, j in enumerate(move):
        diff = (dist[base, j] + dist[i, move]) - (dist[i, j] + dist[base, move])
        ind = np.argmin(diff)
        if diff[ind] < 0:
            is_best = False
            move[ind], move[i] = move[i], move[ind]
    return is_best

def EMD(A, B, n_attempts=3, return_movement=False):
    assert A.shape == B.shape, "Size mismatch"
    n, d = A.shape
    dist = np.sqrt(np.sum(np.square(
        np.reshape(B, (1, n, d)) - np.reshape(A, (n, 1, d)
        )), 2))
    emd = None
    movement = None
    for _ in range(n_attempts):
        move = np.random.choice(n, n, False)
        while not best(dist, move): pass
        new_emd = sum([dist[i,j] for i, j in enumerate(move)]) / n
        if emd is None or new_emd < emd:
            emd = new_emd
            movement = move
    if return_movement:
        return emd, movement
    return emd


# -*- coding: utf-8 -*-
"""
Classes per codificar

@author: david
"""

import numpy as np

class EncoderNone:
    def __init__(self):
        self.size = 1
    
    def encode(self, data):
        return data
    
    def decode(self, X):
        return X
    
    def toNan(self, data):
        return np.isnan(data)

class EncoderDiscrete:
    def __init__(self, minimum=None, maximum=None):
        self.size = 1
        self.has_min = minimum is not None
        self.has_max = maximum is not None
        self.min = np.nan if minimum is None else minimum
        self.max = np.nan if maximum is None else maximum
    
    def encode(self, data):
        return data.astype(float)
    
    def decode(self, X):
        X = np.round(X).astype(int)
        if self.has_min:
            X[X < self.min] = self.min
        if self.has_max:
            X[X > self.max] = self.max
        return X
    
    def toNan(self, data):
        return np.isnan(data)
    
class EncoderLimit:
    def __init__(self, lower=None, upper=None, tails=True, influence=1):
        self.size = 1
        self.tails = tails
        assert lower is not None or upper is not None, "At least one limit must exist"
        self.invert = lower is None
        self.single = lower is None or upper is None
        self.a = upper if self.invert else lower 
        self.range = influence / 2 if self.single else (upper - lower) / 2
        self.sign = -1 if self.invert else 1
    
    def nan_out(self, X):
        tol = 0 if self.tails else 1e-6
        out = (-X if self.single else np.abs(X)) >= 1 + tol
        X[out] = np.nan
        return X
    
    def encode(self, data):
        X = self.nan_out(self.sign * (data - self.a) / self.range - 1)
        if self.tails and self.single:
            X = np.log(np.exp(X + 1) - 1)
        elif self.tails:
            X = np.log((1 + X) / (1 - X))
        return X
    
    def decode(self, X):
        if self.tails and self.single:
            X = np.log(np.exp(X) + 1) - 1
        elif self.tails:
            eX = np.exp(X)
            X = 2 * eX / (1 + eX) - 1
        else:
            X[X > 1] = 1
            X[X < -1] = -1
        data = self.a + self.sign * (X + 1) * self.range
        return data
    
    def toNan(self, data):
        return np.isnan(self.nan_out(self.sign * (data - self.a) / self.range - 1)) 

class EncoderIgnore:
    def __init__(self, default=None):
        self.size = 0
        self.default = default
    
    def encode(self, data):
        return np.zeros((data.shape[0], 0))
    
    def decode(self, X):
        return np.full((X.shape[0], 1), self.default)
    
    def toNan(self, data):
        return np.full(data.shape, False)
    
class EncoderOHE:
    def __init__(self, symbols):
        try:
            symbols = np.array(symbols, dtype=object)
            self.size = len(symbols)
            self.symbols = symbols
        except:
            if type(symbols) is int:
                self.size = symbols
                self.symbols = None
            else:
                raise ValueError("Invalid symbols type")
    
    def encode(self, data):
        if self.symbols is None:
            self.symbols = np.array(list(data.unique()) + [None for _ in range(self.size)], dtype=object)[:self.size] 
        X = np.array([data == symbol for symbol in self.symbols], dtype=float).transpose()
        X[np.sum(X, 2) == 0] = np.full(self.size, np.nan)
        return X
    
    def decode(self, X):
        return np.reshape(self.symbols[np.argmax(X, 1)], (-1, 1))
    
    def toNan(self, data):
        valid = np.vectorize(lambda x: x in self.symbols)
        return ~valid(data)

class EncoderEquivalence:
    def __init__(self, symbols):
        symbols = np.array(symbols, dtype=object)
        self.size = 1
        self.symbols = len(symbols)
        assert self.symbols >= 1, "There must be at least one value"
        values = np.arange(self.symbols)
        self.forward = {symb: val for symb, val in zip(symbols, values)}
        self.backward = {val: symb for symb, val in zip(symbols, values)}
    
    def encode(self, data):
        keys, val = np.unique(data, return_inverse=True)
        X = np.zeros(data.shape)
        for i, key in enumerate(keys):
            X[val == i] = self.forward.get(key, np.nan)
        return X
    
    def decode(self, X):
        X = np.round(X).astype(int)
        X[X < 0] = 0
        X[X >= self.symbols] = self.symbols - 1
        X = np.round(X).astype(int)
        data = np.full(X.shape, self.backward[0], object)
        for val in range(1, self.symbols):
            data[X == val] = self.backward.get(val)
        return data
    
    def toNan(self, data):
        valid = np.vectorize(lambda x: x in self.forward.keys())
        return ~valid(data)



# -*- coding: utf-8 -*-
"""
Transformer

@author: david
"""

import pandas as pd
import numpy as np

class Transformer():
    def __init__(self, normalize=True, remove_cov=True, whitening=False):
        super().__init__()
        self.encoders = dict()
        self.dtypes = dict()
        self.labels = []
        remove_cov = normalize and remove_cov
        whitening = remove_cov and whitening
        self.transforms = {
            "normalize": normalize,
            "remove_cov": remove_cov,
            "whitening": whitening
        }
    
    def add_encoders(self, encoders):
        self.encoders |= encoders
    
    def set_encoder(self, label, encoder):
        self.encoders[label] = encoder
        
    def get_info(self, data):
        self.labels = list(data.columns)
        self.dtypes = self.dtypes
        
    def copy_transformer(self):
        new_trans = Transformer()
        new_trans.encoders = self.encoders.copy()
        new_trans.labels = self.labels.copy()
        new_trans.dtypes = self.dtypes.copy()
        new_trans.transforms = self.transforms.copy()
        return new_trans
        
    def transform(self, data, refit=False, process=True):
        self.labels = list(data.columns)
        self.dtypes = data.dtypes.to_dict() | self.dtypes
        Xfeatures = sum([enc.size for enc in self.encoders.values()])
        X = np.zeros((data.shape[0], Xfeatures))
        i = 0
        for label in self.labels:
            size = self.encoders[label].size
            X[:,i:i + size] = self.encoders[label].encode(data[[label]])
            i += size
        if refit:
            self._fitprocess(X[~np.isnan(X).any(1)])
        if process:
            X = self._preprocess(X)
        return X
    
    def inv_transform(self, X, process=True):
        Xfeatures = sum([enc.size for enc in self.encoders.values()])
        if process:
            X = self._postprocess(X)
        assert X.shape[1] == Xfeatures, f"X n_features ({X.shape[1]}) different from expected ({sum([enc.size for enc in self.encoders.values()])})"
        data = pd.DataFrame(columns=self.labels, dtype=object)
        i = 0
        for label in self.labels:
            size = self.encoders[label].size
            data[[label]] = self.encoders[label].decode(X[:,i:i + size])
            i += size
        return data.astype(self.dtypes)
    
    def covariance(X, treshold):
        _, dim = X.shape
        nums = np.arange(dim)
        cov = np.cov(X.T)
        sigma = np.sqrt(np.var(X, 0))
        sigma_1 = 1 / (np.maximum(sigma, 0) + 1e-6)
        auto_cov = sigma_1 * (sigma_1 * cov).T
        related = treshold < np.abs(auto_cov)
        related[nums, nums] = True
        for i, row in enumerate(related):
            conn = np.any(related[:, row], 1)
            row = np.sum(related[conn], 0, dtype=bool)
            related[conn] = False
            related[i] = row
        blocks = [nums[rel] for rel in related[related.any(1)]]
        signs = np.zeros(dim)
        for block in blocks:
            signs[block] = np.sign(auto_cov[block[0], block])
        return blocks, signs
    
    def _fitprocess(self, X):
        if self.transforms['normalize']:
            self.mu = np.mean(X, 0)
            self.sigma = np.sqrt(np.var(X, 0))
            self.sigma = np.maximum(self.sigma, 0) + 1e-6
            Y = (X - self.mu) / self.sigma
        if self.transforms['remove_cov']:
            self.blocks, self.sign = Transformer.covariance(Y, 0.975)
            self.dim_low, self.dim_high = len(self.blocks), Y.shape[1]
            Y = Y * self.sign
            Z = np.zeros((Y.shape[0], self.dim_low))
            for i, block in enumerate(self.blocks):
                Z[:,i] = np.nanmean(Y[:,block], 1)
        if self.transforms['whitening']:
            lambdas, self.U = np.linalg.eigh(np.cov(Z.T))
            self.s_lambdas = np.sqrt(np.maximum(lambdas, 0)) + 1e-6
            # W = Z @ self.U @ np.diag(1 / self.s_lambdas) @ self.U.T
        
    def row_nanmean(X):
        bad = np.all(np.isnan(X), 1)
        out = np.full(bad.shape, np.nan, dtype=float)
        out[~bad] = np.nanmean(X[~bad], 1)
        return out
        
    def _preprocess(self, X):
        if self.transforms['normalize']:
            X = (X - self.mu) / self.sigma
        if self.transforms['remove_cov']:
            X = X * self.sign
            Y = np.zeros((X.shape[0], self.dim_low))
            for i, block in enumerate(self.blocks):
                Y[:,i] = Transformer.row_nanmean(X[:,block])
            X = Y
        if self.transforms['whitening']:
            X = X @ self.U @ np.diag(1 / self.s_lambdas) @ self.U.T
        return X
        
    def _postprocess(self, X):
        if self.transforms['whitening']:
            X = X @ self.U @ np.diag(self.s_lambdas) @ self.U.T
        if self.transforms['remove_cov']:
            Y = np.zeros((X.shape[0], self.dim_high))
            for i, block in enumerate(self.blocks):
                Y[:,block] = X[:,[i]]
            X = Y * self.sign
        if self.transforms['normalize']:
            X = X * self.sigma + self.mu
        return X
    
    def deformation(self):
        deformation = 1
        if self.transforms['normalize']:
            deformation /= np.prod(self.sigma)
        if self.transforms['remove_cov']:
            ...
        if self.transforms['whitening']:
            deformation /= np.prod(self.s_lambdas)
        return deformation
    
    def run(self, data, FUN):
        nans = self.toNan(data)
        if nans.all():
            raise ValueError("All rows contain nan")
        X = self.transform(data[~nans], refit=True)
        FUN(X)
        return nans
    
    def toNan(self, data):
        nans = np.full(data.shape[0], False)
        for label in self.labels:
            nans |= self.encoders[label].toNan(data[label])
        return nans


# -*- coding: utf-8 -*-
"""
Gaussian Mixture Model

@author: david
"""

import numpy as np
from time import time as gettime

from ._base import BaseGenerator

class GMM(BaseGenerator):
    def __init__(self, k='auto', k_max=None, batches=False, alpha=10, extra_iter=False, **kwargs):
        super().__init__(**kwargs)
        self.k_max = k_max
        if k == 'auto':
            self.k_mode = k
            self.k = None
        elif type(k) is int:
            self.k_mode = 'set'
            self.k = max(k, 1)
        else:
            raise ValueError("k must be an integer or 'auto'")
        self.batches = batches
        self.alpha = alpha
        self.extra_iter = batches and extra_iter
        
    def reset(self):
        self.weights = np.zeros(self.k) + 1 / self.k
        self.means = np.random.normal(size=(self.k, self.dim))
        self.covariances = np.array([np.identity(self.dim) for _ in range(self.k)])
        self.inverses = np.array([np.identity(self.dim) for _ in range(self.k)])
        self.determinants = np.ones(self.k)
        
    def set_k(self, k='auto', k_max=None):
        self.k_max = k_max
        if k == 'auto':
            self.k_mode = k
            self.k = None
        elif type(k) is int:
            self.k_mode = 'set'
            self.k = max(k, 1)
        else:
            raise ValueError("k must be an integer or 'auto'")
    
    def k_probability(self, X):
        assert X.shape[1] == self.dim, "Size mismatch"
        k_prob = []
        for weight, mean, inverse, determinant in \
            zip(self.weights, self.means, self.inverses, self.determinants):
                centered = X - mean
                exponent = -np.sum(centered.dot(inverse) * centered, 1) / 2
                k_prob.append(weight * np.exp(exponent) / np.sqrt(determinant))
                    
        k_prob = np.array(k_prob).transpose() / np.power(2 * np.pi, self.dim / 2)
        return np.nan_to_num(k_prob)
        
    
    def probabilities(self, X):
        return np.sum(self.k_probability(X), 1)
    
    def iterate(self, X):
        # Expectation
        probs = self.k_probability(X).transpose()
        resps = probs / np.maximum(np.sum(probs, 0), 1e-300)
        t_resps = np.maximum(np.sum(resps, 1), 1e-300)
        
        # Maximization
        self.new_weights = t_resps / self.n
        self.new_weights /= np.sum(self.new_weights)
        self.new_means = np.transpose(resps.dot(X).T / t_resps)
        self.new_covariances = np.array([
            ((X - mean).T * resp) @ (X - mean) / t_resp 
            for t_resp, resp, mean in 
                zip(t_resps, resps, self.new_means)
                ])
    
    def step(self, delta):
        self.weights = (1 - delta) * self.weights + delta * self.new_weights
        self.means = (1 - delta) * self.means + delta * self.new_means
        self.covariances = (1 - delta) * self.covariances + delta * self.new_covariances
        diag = np.arange(self.dim)
        for i, covariance in enumerate(self.covariances):
            covariance[diag, diag] += delta * 1e-6
            self.inverses[i] = np.linalg.inv(covariance)
            self.determinants[i] = np.linalg.det(covariance)
        
    def full_iteration(self):
        self.iterate(self.X)
        self.step(1)
        
    def batch_iteration(self):
        if self.n > self.k * self.alpha:
            n, self.n = self.n, self.k * self.alpha
            self.iterate(self.X[np.random.choice(n, self.n, False)])
            self.n = n
            self.step(self.k * self.alpha / self.n)
        else:
            self.full_iteration()
    
    def _fit(self, X, max_iter=1000, max_time=1, llh_tol=1e-3, n_attempts=3):
        self.X = X
        self.n, self.dim = X.shape
        if self.k_mode == 'auto':
            best_bic = None
            best_k = None
            gmm = GMM(batches=self.batches, alpha=self.alpha, \
                      fit_max_iter=max_iter, fit_max_time=max_time, \
                      fit_llh_tol=llh_tol, fit_n_attempts=n_attempts)
            k = 1
            k_lim = 2
            loop = k_lim if self.k_max is None else self.k_max
            while loop != 0:
                #print(k)
                loop -= 1
                gmm.set_k(k)
                gmm.fit(X)
                bic = k * (self.dim * (self.dim + 1) / 2 + self.dim + 1) * np.log(self.n) \
                    - 2 * gmm.loglikelihood(X)
                if best_bic is None or bic < best_bic:
                    best_bic = bic
                    best_k = k
                    if self.k_max is None:
                        loop = k_lim
                k += 1
            self.k = best_k
            
        best_llh = None
        for attempt in range(n_attempts):
            self.reset()
            loop = True
            iters = max_iter
            time = max_time + gettime()
            llh = self.loglikelihood(X)
            while loop:
                if self.batches:
                    self.batch_iteration()
                else:
                    self.full_iteration()
                    
                iters -= 1
                last_llh = llh
                llh = self.loglikelihood(X)
                
                loop &= iters > 0
                loop &= llh - last_llh > llh_tol
                loop &= time > gettime()
                
            if not best_llh or llh > best_llh:
                best_llh = llh
                best_params = (self.weights, self.means, self.covariances, self.inverses, self.determinants)
        self.weights, self.means, self.covariances, self.inverses, self.determinants = best_params
        if self.extra_iter:
            self.full_iteration()
            
    def _generate(self, size):
        ind, num = np.unique(np.random.choice(self.k, size, p=self.weights), return_counts=True)
        s = [
            np.random.multivariate_normal(self.means[i], self.covariances[i], n)
            for i, n in zip(ind, num)
            ]
        S = np.concatenate(s)
        return S
    
    def _fill(self, Y):
        assert Y.shape[1] == self.dim, "Size mismatch"
        for y in Y:
            bad = np.isnan(y)
            good = np.logical_not(bad)
            goods = np.sum(good)
            k_prob = []
            tol = 1e-6
            for weight, mean, covariance in \
                zip(self.weights, self.means, self.covariances):
                    centered = (y - mean)[good]
                    _covariance = covariance[good][:,good] + tol * np.identity(goods)
                    determinant = np.linalg.det(_covariance)
                    inverse = np.linalg.inv(_covariance)
                    exponent = (centered @ inverse @ centered) / 2
                    k_prob.append(weight * np.exp(-exponent) / np.sqrt(determinant))
            prob = np.array(k_prob)
            ind = np.random.choice(self.k, p=prob / np.sum(prob))
            covariance = self.covariances[ind]
            inv_subcov = np.linalg.inv(covariance[good][:,good] + tol * np.identity(goods))
            new_mean = self.means[ind][bad] + covariance[bad][:,good] @ inv_subcov @ (y - self.means[ind])[good]
            new_cov = covariance[bad][:,bad] - covariance[bad][:,good] @ inv_subcov @ covariance[good][:,bad]
            y[bad] = np.random.multivariate_normal(new_mean, new_cov)
        return Y



# -*- coding: utf-8 -*-
"""
Variational AutoEncoder

@author: david
"""

import numpy as np
import torch
import optuna
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from time import time as gettime

from ._base import BaseGenerator

def default_loss_function(recon_x, x, mu, logvar):
    recLoss = nn.MSELoss(reduction='sum')
    KLD = 0.5 * torch.sum(logvar.exp() + mu.pow(2) - 1 - logvar)
    return recLoss(recon_x, x) + torch.sqrt(KLD)

class VAE(BaseGenerator, nn.Module):
    def __init__(self, device='auto', **kwargs):
        super().__init__(**kwargs)
        if device == 'auto':
            self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        elif type(device) == str:
            self._device = torch.device(device)
        else:
            self._device = device
        
    def encode(self, x):
        enc = self.encoder(x)
        return self.mean(enc), self.logvar(enc)

    def reparametrize(self, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.zeros(std.size(), device=self._device.type, dtype=torch.float).normal_()
        return mean + eps * std

    def decode(self, z):
        dec = self.decoder(z)
        return dec

    def forward(self, x):
        mean, logvar = self.encode(x)
        z = self.reparametrize(mean, logvar)
        return self.decode(z), mean, logvar
    
    def _train(self, X, epochs, criterion=default_loss_function, batches=None, batch_size=128,
               lr=1e-1, reduce_lr=True, min_lr=1e-5, patience='auto', max_time=10):
        if patience == 'auto':
            patience = 5 if reduce_lr else 20
        loop = True
        epoch = 0
        best = None
        since = 0
        time = max_time + gettime()
        batches = max(round(self.n / batch_size), 1) if batches is None else batches
        batch_size = int(np.ceil(self.n / batches))
        train_loader = torch.utils.data.DataLoader(X, batch_size=batch_size, shuffle=True)
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        schedule = ReduceLROnPlateau(optimizer, patience=patience)
        self.train()
        while loop:
            loss = 0
            for _batch in train_loader:
                batch = _batch
                optimizer.zero_grad()
                
                outputs, mean, logvar = self.forward(batch)
                
                train_loss = criterion(outputs, batch, mean, logvar)
                train_loss.backward()
                
                optimizer.step()
                
                loss += train_loss.detach().cpu()
            if reduce_lr:
                schedule.step(loss)
                
            if not epochs == 'auto':
                epoch += 1
                loop = epoch < epochs
            elif reduce_lr:
                loop = optimizer.state_dict()['param_groups'][0]['lr'] > min_lr
            else:
                since += 1
                if best is None or best > loss:
                    best = loss
                    since = 0
                loop = since < patience
            loop &= time > gettime()
        self.eval()
    
    def setup_run(self, X, phases, **args):
        self.enc_dim = phases[-1]
        
        f = nn.Sigmoid()
        recLoss = nn.MSELoss(reduction='sum')
        
        self.encoder = nn.Sequential(*[
            f if i == -1 else nn.Linear(phases[i], phases[i + 1])
            for i in [e // 2 if e % 2 == 0 else -1 for e in range(2 * len(phases) - 4)]
            ])
        self.mean = nn.Linear(phases[-2], self.enc_dim)
        self.logvar = nn.Linear(phases[-2], self.enc_dim)
        self.decoder = nn.Sequential(*[
            f if i == -1 else nn.Linear(phases[i], phases[i - 1])
            for i in [e // 2 if e % 2 == 0 else -1 for e in range(2 * len(phases) - 2, 1, -1)]
            ])
        
        self.to(self._device)
        self._train(X, **args)
        with torch.no_grad():
            Y, m, l = self.forward(X)
        return recLoss(X, Y)
    
    def prefit(self, X, **prefit_args):
        self._prefit(X, **self.fit_args | prefit_args)
    
    def _prefit(self, X, enc_dim='auto', layers='auto', epochs='auto', auto_trials=8, **train_args):
        tX = torch.from_numpy(X).float().to(self._device)
        self.n, self.dim = X.shape
        self._optimze_args = {'enc_dim': enc_dim, 'layers': layers}
        assert enc_dim == 'auto' or layers == 'auto'
        
        def objective(trial):
            enc_dim = self._optimze_args['enc_dim']
            layers = self._optimze_args['layers']
            enc_dim = trial.suggest_int('enc_dim', 1, self.dim) if enc_dim == 'auto' else enc_dim
            layers = trial.suggest_int('layers', 1, 5) if layers == 'auto' else layers
            
            phases = torch.tensor(np.round(np.exp(np.linspace(np.log(self.dim), np.log(enc_dim), layers + 1))), dtype=int)
            
            return self.setup_run(tX, phases, epochs=epochs, **train_args)
        
        optuna.logging.set_verbosity(optuna.logging.ERROR)
        study = optuna.create_study()
        study.optimize(objective, n_trials=auto_trials)
        
        self.fit_args |= study.best_params
        self.encoder = None
        self.mean = None
        self.logvar = None
        self.decoder = None
    
    def _fit(self, X, phases='auto', enc_dim='auto', layers='auto', epochs='auto', auto_trials=8, **train_args):
        tX = torch.from_numpy(X).float().to(self._device)
        self.n, self.dim = X.shape
        
        if phases == 'auto':
            self._optimize_args = {'enc_dim': enc_dim, 'layers': layers}
            
            if enc_dim == 'auto' or layers == 'auto':
                def objective(trial):
                    enc_dim = self._optimize_args['enc_dim']
                    layers = self._optimize_args['layers']
                    enc_dim = trial.suggest_int('enc_dim', 1, self.dim) if enc_dim == 'auto' else enc_dim
                    layers = trial.suggest_int('layers', 1, 5) if layers == 'auto' else layers
                    
                    phases = torch.tensor(np.round(np.exp(np.linspace(np.log(self.dim), np.log(enc_dim), layers + 1))), dtype=int)
                    
                    return self.setup_run(tX, phases, epochs=epochs, **train_args)
                
                optuna.logging.set_verbosity(optuna.logging.ERROR)
                study = optuna.create_study()
                study.optimize(objective, n_trials=auto_trials)
                
                self._optimize_args |= study.best_params
            enc_dim = self._optimize_args['enc_dim']
            layers = self._optimize_args['layers']
            
            phases = np.array(np.round(np.exp(np.linspace(np.log(self.dim), np.log(enc_dim), layers + 1))), dtype=int)
            
        return self.setup_run(tX, phases, epochs=epochs, **train_args)
            
    def _generate(self, size):
        with torch.no_grad():
            S = torch.zeros((size, self.enc_dim), device=self._device, dtype=torch.float).normal_()
            X = self.decode(S)
        return X.detach().cpu().numpy().astype(float)


# -*- coding: utf-8 -*-
"""
Base Generator

@author: david
"""

import numpy as np

class BaseGenerator():
    def __init__(self, **kwargs):
        super().__init__()
        self.fitted = False
        self.fit_args = dict()
        self.gen_args = dict()
        self.fill_args = dict()
        self.set_args(**kwargs)
        
    def set_args(self, **kwargs):
        fit = ('fit', self.set_fit_args)
        gen =  ('gen', self.set_generate_args)
        fill =  ('fill', self.set_fill_args)
        for code, f in (fit, gen, fill):
            f(**{k[len(code)+1:]: v for k, v in kwargs.items() if k[:len(code)+1] == (code + '_')})
        
    def set_fit_args(self, **kwargs):
        self.fit_args = self.fit_args | kwargs
        
    def set_generate_args(self, **kwargs):
        self.gen_args = self.gen_args | kwargs
        
    def set_fill_args(self, **kwargs):
        self.fill_args = self.fill_args | kwargs
        
    def loglikelihood(self, X):
        assert self.fitted
        return np.sum(np.log(self.probabilities(X) + 1e-300))
    
    def fit(self, X, **kwargs):
        self.fitted = True
        out = self._fit(X, **self.fit_args | kwargs)
        return out
    
    def generate(self, size, **kwargs):
        assert self.fitted
        return self._generate(size, **self.gen_args | kwargs)
    
    def fill(self, X, **kwargs):
        assert self.fitted
        return self._fill(X, **self.fill_args | kwargs)
    


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 10 18:37:51 2022

@author: david
"""

import numpy as np

from ._classes import EncoderNone, EncoderDiscrete, EncoderIgnore, EncoderOHE, EncoderLimit, EncoderEquivalence
    
def greater(value=0, include=True, influence=1):
    if include:
        return EncoderLimit(lower=value, tails=False)
    else:
        return EncoderLimit(lower=value, tails=True, influence=influence)
    
def lower(value=0, include=True, influence=1):
    if include:
        return EncoderLimit(upper=value, tails=False)
    else:
        return EncoderLimit(upper=value, tails=True, influence=influence)

def auto(data):
    if data.dtype == object:
        symbols = data.unique()
        n = len(data)
        s = len(symbols)
        if s == 1:
            return EncoderIgnore(symbols[0])
        elif s == 2:
            return EncoderEquivalence(symbols)
        elif np.sqrt(n) > s and s < 100:
            return EncoderOHE(list(symbols))
        else:
            return EncoderIgnore('Ignored')
    else:
        if data.dtype == np.float64:
            if (data > 1e-6).all():
                return EncoderLimit(lower=0, tails=True)
            elif (data >= -0).all():
                return EncoderLimit(lower=0, tails=False)
            else:
                return EncoderNone()
        else:
            minimum = None
            if (data > 0).all():
                minimum = 1
            elif (data >= 0).all():
                minimum = 0
            return EncoderDiscrete(minimum=minimum)


# -*- coding: utf-8 -*-
"""
DataHub

@author: david
"""

import numpy as np
import pandas as pd
import multiprocessing as mp
import dill
from time import time as gettime

import synthdata.encoder as enc
from ._transformer import Transformer

class DataHub(Transformer):
    def __init__(self, cores=None, **kwrds):
        super().__init__(**kwrds)
        if not cores:
            cores = 1
        self.use_mp = (cores != 1)
        self.cores = cores
    
    def load(self, data, encoders=dict(), method=None):
        assert len(data.shape) == 2, "data must be a 2d array of shape (n_samples, n_features)"
        
        self.data = data
        self.get_info(data)
        self.add_encoders({
            label: enc.auto(data[label])
            for label in self.labels
            } | encoders)
        self.method = None
        self.samples, self.features = data.shape
        
    def set_method(self, method):
        self.method = method
        
    def set_cores(self, cores=None):
        if not cores:
            cores = 1
        self.use_mp = (cores != 1)
        self.cores = cores
    
    MP_DATA_FILE = 'temp.pickle'
    
    def pickle_data(self):
        class PickleManager:
            def __enter__(*_):
                self.data.to_pickle(DataHub.MP_DATA_FILE)
                return DataHub.MP_DATA_FILE
            
            def __exit__(*_):
                import os
                os.remove(DataHub.MP_DATA_FILE)
                return False
            
        return PickleManager()
    
    def _run_mp(what):
        FUN, trans, (target, key), args, kwds = dill.loads(what)
        data = pd.read_pickle(DataHub.MP_DATA_FILE)
        subdata = data[data[target] == key]
        return FUN(trans, subdata, *args, **kwds)
    
    def for_target(self, target, FUN, *args, **kwargs):
        if target is None:
            results = {'all': FUN(self, self.data, *args, **kwargs)}
        else:
            target_encoder = self.encoders[target]
            results = dict()
            if self.use_mp:
                keys = self.data[target].unique()
                values = []
                neccesary_cores = len(keys)
                allowed_cores = self.cores if self.cores > 0 else (mp.cpu_count() + self.cores)
                cores = min(neccesary_cores, allowed_cores)
                with self.pickle_data(), mp.Pool(cores) as pool:
                    for key in keys:
                        self.encoders[target] = enc.ignore(default=key)
                        trans = self.copy_transformer()
                        what = [dill.dumps((FUN, trans, (target, key), args, kwargs))]
                        values.append(pool.apply_async(DataHub._run_mp, what))
                    for key, value in zip(keys, values):
                        results[str(key)] = value.get()
            else:
                for value in self.data[target].unique():
                    self.encoders[target] = enc.ignore(default=value)
                    subdata = self.data[self.data[target] == value]
                    results[str(value)] = FUN(self, subdata, *args, **kwargs)
            self.encoders[target] = target_encoder
        return results
    
    def _kfold(trans, subdata, method, n_samples, folds, validation, return_fit, return_time):
        def sample_fold_total(n_samples, folds, total):
            if folds is None:
                assert total >= 2 * n_samples, "n_samples to big for the data"
                folds = total // n_samples
                return n_samples, folds, n_samples * folds
            elif n_samples is None:
                assert folds >= 2, "folds must be at least 2"
                n_samples = total // folds
                return n_samples, folds, folds * n_samples
            else:
                assert total >= 2 * n_samples, "n_samples to big for the data"
                folds = min(folds, total // n_samples)
                return n_samples, folds, n_samples * folds
        
        subsample = len(subdata)
        sample, fold, total = sample_fold_total(n_samples, folds, subsample)
        sampling = np.reshape(np.random.choice(subsample, total, False), (fold, -1))
        value = 0
        selfvalue = 0
        fit_time = 0
        eval_time = 0
        for i in range(fold):
            train = subdata.iloc[sampling[i]]
            test = subdata.iloc[sampling[(i + 1) % fold]]
            start = gettime()
            trans.run(train, method.fit)
            fit_time += gettime() - start
            if validation == 'loglikelihood':
                if return_fit:
                    selfvalue += method.loglikelihood(trans.transform(train)) / sample + np.log(trans.deformation())
                start = gettime()
                value += method.loglikelihood(trans.transform(test)) / sample + np.log(trans.deformation())
                eval_time += gettime() - start
            else:
                start = gettime()
                Xgen = trans.transform(trans.inv_transform(method.generate(sample)), process=False)
                eval_time += gettime() - start
                Xtrain = trans.transform(train, process=False)
                Xtest = trans.transform(test, process=False)
                if return_fit:
                    selfvalue += validation(Xtrain, Xgen)
                value += validation(Xtest, Xgen)
        output = {'validation': value / fold}
        if return_fit:
            output |= {'train': selfvalue / fold}
        if return_time:
            output |= {'fitting_time': fit_time / fold, 'evaluation_time': eval_time / fold}
        return output
    
    def kfold_validation(self, n_samples=None, folds=None, method=None, validation='loglikelihood', target=None, return_fit=False, return_time=True):
        method = self.method if method is None else method
        if n_samples is None and folds is None:
            raise ValueError("No value specified for kfold validation")
        args = {
            'method': method,
            'n_samples': n_samples,
            'folds': folds,
            'validation': validation,
            'return_fit': return_fit,
            'return_time': return_time
        }
        if target is None:
            return self.for_target(target, DataHub._kfold, **args)['all']
        return self.for_target(target, DataHub._kfold, **args)

    def _generate(trans, subdata, method, n_samples):
        trans.run(subdata, method.fit)
        return trans.inv_transform(method.generate(n_samples))
    
    def generate(self, n_samples, method=None, target=None):
        method = self.method if method is None else method
        
        output = self.for_target(target, DataHub._generate, method=method, n_samples=n_samples)
        return pd.concat(output.values(), ignore_index=True)
        
    def _fill(trans, subdata, method):
        new_data = subdata.copy()
        nans = trans.run(new_data, method.fit)
        new_data.iloc[nans] = trans.inv_transform(
            method.fill(trans.transform(new_data[nans]))
        )
        return new_data
    
    def fill(self, method=None, target=None):
        method = self.method if method is None else method
        
        output = self.for_target(target, DataHub._fill, method)
        return pd.concat(output.values(), ignore_index=True)
     
    def _extend(trans, subdata, method, n_samples, max_samples):
        subsample = len(subdata)
        if n_samples <= subsample:
            return subdata.iloc[np.random.choice(subsample, min(max_samples, subsample), False)]
        else:
            trans.run(subdata, method.fit)
            new_data = trans.inv_transform(method.generate(n_samples - subsample))
            return pd.concat([subdata, new_data])
    
    def extend(self, n_samples, max_samples='n_samples', method=None, target=None):
        method = self.method if method is None else method
        if max_samples == 'n_samples':
            max_samples = n_samples
        elif max_samples is None:
            max_samples = self.samples
        else:
            max_samples = max(max_samples, n_samples)
            
        output = self.for_target(target, DataHub._extend, method, n_samples, max_samples)
        return pd.concat(output.values(), ignore_index=True)
